package org.opennms.netmgt.jasper.compiler;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import net.sf.jasperreports.engine.JRException;
import net.sf.jasperreports.engine.JRParameter;
import net.sf.jasperreports.engine.JRSubreport;
import net.sf.jasperreports.engine.JasperCompileManager;
import net.sf.jasperreports.engine.JasperReport;
import net.sf.jasperreports.engine.util.JRElementsVisitor;
import net.sf.jasperreports.engine.util.JRVisitorSupport;

import org.apache.commons.io.FileUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.vaadin.Application;
import com.vaadin.data.Item;
import com.vaadin.data.Property;
import com.vaadin.data.Property.ValueChangeEvent;
import com.vaadin.data.util.IndexedContainer;
import com.vaadin.event.FieldEvents.TextChangeEvent;
import com.vaadin.event.FieldEvents.TextChangeListener;
import com.vaadin.ui.AbstractTextField.TextChangeEventMode;
import com.vaadin.ui.Button;
import com.vaadin.ui.Button.ClickEvent;
import com.vaadin.ui.Button.ClickListener;
import com.vaadin.ui.HorizontalLayout;
import com.vaadin.ui.OptionGroup;
import com.vaadin.ui.PopupView;
import com.vaadin.ui.Table;
import com.vaadin.ui.TextField;
import com.vaadin.ui.Window;
import com.vaadin.ui.Window.Notification;

import edu.emory.mathcs.backport.java.util.Arrays;
import edu.emory.mathcs.backport.java.util.Collections;

public class JasperCompilerApplication extends Application implements Property.ValueChangeListener {
    
    private class SubReportInfo{
        
        private static final String DESTINATION_DIR = "target/reports/subreports/";
        private static final String SOURCE_DIR = "src/main/reports/subreports/";
        private String m_jasperFilename;
        private String m_jrxmlFilename;
        
        public SubReportInfo(String reportFilename) {
            m_jasperFilename = reportFilename;
            m_jrxmlFilename = reportFilename.replaceAll(".jasper", ".jrxml");
        }

        public String getSourceJRXML() {
            return SOURCE_DIR + m_jrxmlFilename;
        }

        public String getOutputJasper() {
            return DESTINATION_DIR + m_jasperFilename;
        }

        public String getSourceFilename() {
            return m_jrxmlFilename;
        }

        public String getSourceJRXMLDestination() {
            return DESTINATION_DIR + m_jrxmlFilename;
        }

        public String getSubreportDestination() {
            return DESTINATION_DIR;
        }
    }
    
    private static final String REPORT_PROPERTY_NAME = "name";
    
    private Window m_window;
    private Pattern m_filenamePattern = Pattern.compile("\"(.*?.jasper)");
    private Pattern m_subreportProperty = Pattern.compile("\\$P\\{(.*?)\\}");
    private Table m_table;
    private PopupView m_loadingWindow;
    private Window m_configWindow;
    private OptionGroup m_configOptions;
    private String m_installDir = "";
    private String m_currentProfile = "";
    private Map<String, String> m_profiles = new HashMap<String, String>();
    
    @Override
    public void init() {
        
        m_window = new Window("Jasper Compiler App");
        setMainWindow(m_window);
        
        m_configOptions = new OptionGroup("Select Profile", getProfiles());
        m_configOptions.addListener(this);
        Button configBtn = new Button("OK");
        configBtn.addListener(new ClickListener() {
            
            public void buttonClick(ClickEvent event) {
                m_configWindow.getParent().removeWindow(m_configWindow);
                getMainWindow().showNotification(String.format("You have selected profile: %s with install.dir: %s", m_currentProfile, m_installDir));
            }
        });
        
        m_configWindow = new Window("Profile Selection");
        m_configWindow.setModal(true);
        m_configWindow.addComponent(m_configOptions);
        m_configWindow.addComponent(configBtn);
        m_configWindow.setClosable(false);
        
        getMainWindow().addWindow(m_configWindow);
        
        
        final File file = new File("src/main/reports");
        
        String[] reports = file.list(new FilenameFilter() {

            public boolean accept(File dir, String filename) {
                if(filename.contains(".jrxml")) {
                    return true;
                }
                return false;
            }
        });
        
        final IndexedContainer dataContainer = new IndexedContainer();
        dataContainer.addContainerProperty(REPORT_PROPERTY_NAME, String.class, "");
        
        m_table = new Table(null, dataContainer);
        m_table.setColumnHeader(REPORT_PROPERTY_NAME, "Report Name");
        m_table.setWidth("100%");
        m_table.setSelectable(true);
        m_table.setMultiSelect(true);
        
        for(String reportName : reports) {
            Object itemId = dataContainer.addItem();
            Item item = dataContainer.getItem(itemId);
            item.getItemProperty("name").setValue(reportName);
            //m_table.addItem(new Object[] {reportName}, null);
        }
        
        TextField filterField = new TextField();
        filterField.setTextChangeEventMode(TextChangeEventMode.LAZY);
        filterField.setTextChangeTimeout(300);
        filterField.addListener(new TextChangeListener() {

            public void textChange(TextChangeEvent event) {
                dataContainer.removeAllContainerFilters();
                dataContainer.addContainerFilter("name", event.getText(), true, false);
            }
            
        });
        m_window.addComponent(filterField);
        m_window.addComponent(m_table);
        
        HorizontalLayout btnLayout = new HorizontalLayout();
        
        Button compileReportBtn = new Button("Compile Selected");
        compileReportBtn.addListener(new Button.ClickListener() {
            
            public void buttonClick(ClickEvent event) {
                Set<Object> ids =  (Set<Object>) m_table.getValue();
                
                for(Object id : ids) {
                    Item report = m_table.getItem(id);
                    try {
                        compileReport( file.getAbsolutePath() + File.separator,(String) report.getItemProperty(REPORT_PROPERTY_NAME).getValue());
                    } catch (IOException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
                }
            }
        });
        
        Button compileAllBtn = new Button("Compile All Reports");
        compileAllBtn.addListener(new Button.ClickListener() {
            
            public void buttonClick(ClickEvent event) {
                m_window.showNotification("Compiling All reports");
            }
        });
        
        btnLayout.addComponent(compileReportBtn);
        btnLayout.addComponent(compileAllBtn);
        m_window.addComponent(btnLayout);
    }
    
    private Collection<?> getProfiles() {
        Map<String, String> profiles;
        try {
            profiles = parsePomProfiles("pom.xml");
            return Arrays.asList(profiles.keySet().toArray());
        } catch (ParserConfigurationException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (SAXException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return Collections.emptyList();
    }

    protected Map<String, String> parsePomProfiles(String pom) throws ParserConfigurationException, SAXException, IOException {
        DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        Document dom = db.parse(pom);
        NodeList nodes = dom.getElementsByTagName("profile");
        
        for(int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            NodeList elems = node.getChildNodes();
            
            String version = "";
            String path = "";
            for(int k = 0; k < elems.getLength(); k++) {
                Node n = elems.item(k);
                if(n.getNodeName().equals("id")){
                    version = n.getTextContent();
                }
                
                if(n.getNodeName().equals("properties")) {
                    String openNMSVersion = n.getChildNodes().item(1).getTextContent(); 
                    path = n.getChildNodes().item(3).getTextContent();
                    path = path.replaceAll("\\$\\{opennmsVersion\\}", openNMSVersion);
                    
                }
            }
            
            m_profiles.put(version, path);
        }
        return m_profiles;
        
    }
    
    protected void compileReport(String path, String report) throws IOException {
        try {
            InputStream filteredReport = getFilteredReport(path + report);
            
            JasperReport jasperReport = JasperCompileManager.compileReport(filteredReport);
            
            String destFileName = "target/reports/" + report.replace(".jrxml", ".jasper");
            new File("target/reports/").mkdirs();
            JasperCompileManager.compileReportToFile(path + report, destFileName);
            
            filteredReport.reset();
            FileUtils.copyInputStreamToFile(filteredReport, new File("target/reports/" + report));
            
            compileSubreports(jasperReport);
            
            Notification note = new Notification(report + " has finished compiling", Notification.TYPE_TRAY_NOTIFICATION);
            note.setDelayMsec(1000);
            
            //m_window.showNotification(note);
        } catch (JRException e) {
            m_window.showNotification("Error Occurred While compiling " + report, Notification.TYPE_ERROR_MESSAGE);
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        
    }

    public InputStream getFilteredReport(String report) throws IOException {
        File file = new File(report);
        String fileStr = FileUtils.readFileToString(file);
        fileStr = fileStr.replaceAll("\\$\\{install.dir\\}", m_installDir);
        fileStr = fileStr.replaceAll("\\$\\{opennms.home\\}", m_installDir);
        
        return new ByteArrayInputStream(fileStr.getBytes());
    }
    
    public void compileSubreports( final JasperReport mainReport){
        JRElementsVisitor.visitReport(mainReport, new JRVisitorSupport() {
        
                    @Override
            public void visitSubreport(JRSubreport subreport) {
                String expression = subreport.getExpression().getText();
                String reportFilename = getSubreportFilename(expression);
                String reportDir = getSubreportDirParameter(expression);
                String filename ="";
                
                
                JRParameter[] parameters = mainReport.getParameters();
                for(JRParameter parameter : parameters) {
                    if(parameter.getName().equals(reportDir)) {
                        filename = parameter.getDefaultValueExpression().getText();
                    }
                }
                
                String sourceJRXML = "src/main/reports/subreports/" + reportFilename.replaceAll(".jasper", ".jrxml");
                
                SubReportInfo reportInfo = new SubReportInfo(reportFilename);
                
                try {
                    compile(reportInfo);
                    
                    JasperReport report = JasperCompileManager.compileReport(sourceJRXML);
                    compileSubreports(report);
                } catch (JRException e) {
                    System.err.println("Exception Occurred for report: " + sourceJRXML);
                    e.printStackTrace();
                } catch (IOException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
                
                
            }
            
        });
        
    }
    
    private void compile(SubReportInfo reportInfo) throws IOException, JRException {
        InputStream filteredReport = getFilteredReport(reportInfo.getSourceJRXML());

        checkForSubreportDir(reportInfo);
        
        File outputFile = new File(reportInfo.getOutputJasper());
        FileOutputStream outputStream = new FileOutputStream(outputFile);
        
        JasperCompileManager.compileReportToStream(filteredReport, outputStream);
        
        filteredReport.reset();
        //Make sure you send in the report name instead of the full name
        copyJRXML(filteredReport, reportInfo);
    }
    
    private void checkForSubreportDir(SubReportInfo reportInfo) {
        File subreportDir = new File(reportInfo.getSubreportDestination());
        if(!subreportDir.exists()) {
            subreportDir.mkdirs();
        }
    }

    private void copyJRXML(InputStream intputStream, SubReportInfo reportInfo) throws IOException {
        File source = new File(reportInfo.getSourceJRXMLDestination());
        FileUtils.copyInputStreamToFile(intputStream, source);
        
    }

    public String getSubreportDirParameter(String expression) {
        Matcher matcher = m_subreportProperty.matcher(expression);
        
        if(matcher.find()) {
            if(matcher.groupCount() > 0) {
                return matcher.group(1);
            }
        }
        return null;
    }
    

    public String getSubreportFilename(String expression) {
        Matcher matcher = m_filenamePattern.matcher(expression);
        
        if(matcher.find()) {
            if(matcher.groupCount() > 0) {
                return matcher.group(1);
            }
            
        }
            return null;
        
    }

    public void valueChange(ValueChangeEvent event) {
        m_currentProfile = (String) event.getProperty().getValue();
        m_installDir = m_profiles.get(m_currentProfile);
    }

}
